"""Sphinx configuration."""

project = "Maison"
author = "Dom Batten"
copyright = "2021, Dom Batten"  # noqa: A001
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx_click",
    "myst_parser",
]
autodoc_typehints = "description"
html_theme = "furo"



"""Module to hold the `IniSource` class definition."""

from configparser import ConfigParser
from functools import lru_cache
from typing import Any
from typing import Dict

from .base_source import BaseSource


class IniSource(BaseSource):
    """Class to represent a `.ini` config source."""

    def to_dict(self) -> Dict[Any, Any]:
        """Convert the source config file to a dict.

        Returns:
            a dict of the config options and values
        """
        config = self._load_file()
        return {section: dict(config.items(section)) for section in config.sections()}

    @lru_cache
    def _load_file(self) -> ConfigParser:
        """Load the `.ini` file.

        Returns:
            a `ConfigParser` object with the `.ini` source read into it
        """
        config = ConfigParser()
        config.read(self.filepath)
        return config



"""Store the classes and fixtures used throughout the tests."""

from pathlib import Path
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional

import pytest
import toml


@pytest.fixture(name="create_tmp_file")
def create_tmp_file_fixture(tmp_path: Path) -> Callable[..., Path]:
    """Fixture for creating a temporary file."""

    def _create_tmp_file(content: str = "", filename: str = "file.txt") -> Path:
        tmp_file = tmp_path / filename
        tmp_file.write_text(content)
        return tmp_file

    return _create_tmp_file


@pytest.fixture(name="create_toml")
def create_toml_fixture(create_tmp_file: Callable[..., Path]) -> Callable[..., Path]:
    """Fixture for creating a `.toml` file."""

    def _create_toml(
        filename: str,
        content: Optional[Dict[str, Any]] = None,
    ) -> Path:
        content = content or {}
        config_toml = toml.dumps(content)
        return create_tmp_file(content=config_toml, filename=filename)

    return _create_toml


@pytest.fixture()
def create_pyproject_toml(create_toml: Callable[..., Path]) -> Callable[..., Path]:
    """Fixture for creating a `pyproject.toml`."""

    def _create_pyproject_toml(
        section_name: str = "foo",
        content: Optional[Dict[str, Any]] = None,
        filename: str = "pyproject.toml",
    ) -> Path:
        content = content or {"bar": "baz"}
        config_dict = {"tool": {section_name: content}}
        return create_toml(filename=filename, content=config_dict)

    return _create_pyproject_toml



"""Tests for the `TomlSource` class."""

import re
from pathlib import Path
from textwrap import dedent
from typing import Callable

import pytest

from maison.config_sources.toml_source import TomlSource
from maison.errors import BadTomlError


class TestToDict:
    """Tests for the `to_dict` method."""

    def test_success(self, create_toml: Callable[..., Path]) -> None:
        """A `.toml` is loaded and converted to a `dict`"""
        toml_path = create_toml(filename="config.toml", content={"foo": "bar"})

        toml_source = TomlSource(filepath=toml_path, package_name="acme")

        assert toml_source.to_dict() == {"foo": "bar"}

    def test_toml_decode_error(self, create_toml: Callable[..., Path]) -> None:
        """Toml decoding errors are reported"""
        toml_path = create_toml(filename="config.toml")
        toml_path.write_text(
            dedent(
                """
                "foo" = "bar"
                "foo" = "bar"
                """
            ),
            encoding="utf-8",
        )

        toml_source = TomlSource(filepath=toml_path, package_name="acme")

        error_regex = re.escape(f"Error trying to load toml file '{toml_path!s}'")
        with pytest.raises(BadTomlError, match=error_regex):
            toml_source.to_dict()



"""Module to hold the `BaseSource` abstract class definition."""

from abc import ABC
from abc import abstractmethod
from pathlib import Path
from typing import Any
from typing import Dict


class BaseSource(ABC):
    """Base class from which concrete source abstractions extend."""

    def __init__(self, filepath: Path, package_name: str) -> None:
        """Initialize the object.

        Args:
            filepath: the `Path` to the config file
            package_name: the name of the package, used to pick out the relevant section
                in a `.toml` or `.ini` file
        """
        self.filepath = filepath
        self.package_name = package_name

    def __repr__(self) -> str:
        """Return the __repr__.

        Returns:
            the representation
        """
        return f"<class '{self.__class__.__name__}' path:{self.filepath}>"

    def __str__(self) -> str:
        """Return the __str__.

        Returns:
            the representation
        """
        return self.__repr__()

    @property
    def filename(self) -> str:
        """Return the filename.

        Returns:
            the filename of the source
        """
        return self.filepath.name

    @abstractmethod
    def to_dict(self) -> Dict[Any, Any]:
        """Convert the source config file to a dict.

        Returns:
            a dict of the config options and values
        """



"""Module to hold various utils."""

from pathlib import Path
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional

from maison.config_sources.base_source import BaseSource
from maison.config_sources.ini_source import IniSource
from maison.config_sources.pyproject_source import PyprojectSource
from maison.config_sources.toml_source import TomlSource


def path_contains_file(path: Path, filename: str) -> bool:
    """Determine whether a file exists in the given path.

    Args:
        path: the path in which to search for the file
        filename: the name of the file

    Returns:
        A boolean to indicate whether the given file exists in the given path
    """
    return (path / filename).is_file()


def get_file_path(
    filename: str, starting_path: Optional[Path] = None
) -> Optional[Path]:
    """Search for a file by traversing up the tree from a path.

    Args:
        filename: the name of the file or an absolute path to a config to search for
        starting_path: an optional path from which to start searching

    Returns:
        The `Path` to the file if it exists or `None` if it doesn't
    """
    filename_path = Path(filename).expanduser()
    if filename_path.is_absolute() and filename_path.is_file():
        return filename_path

    start = starting_path or Path.cwd()

    for path in _generate_search_paths(starting_path=start):
        if path_contains_file(path=path, filename=filename):
            return path / filename

    return None


def _generate_search_paths(starting_path: Path) -> Generator[Path, None, None]:
    """Generate paths from a starting path and traversing up the tree.

    Args:
        starting_path: a starting path to start yielding search paths

    Yields:
        a path from the tree
    """
    yield from [starting_path, *starting_path.parents]


def _collect_configs(
    package_name: str,
    source_files: List[str],
    starting_path: Optional[Path] = None,
) -> List[BaseSource]:
    """Collect configs and return them in a list.

    Args:
        package_name: the name of the package to be used to find the right section in
            the config file
        source_files: a list of source config filenames to look for.
        starting_path: an optional starting path to start the search

    Returns:
        a list of the found config sources
    """
    sources: List[BaseSource] = []

    for source in source_files:
        file_path = get_file_path(
            filename=source,
            starting_path=starting_path,
        )

        if not file_path:
            continue

        # Dict[str, Any] to stop mypy complaining:
        # https://github.com/python/mypy/issues/5382#issuecomment-583901369
        source_kwargs: Dict[str, Any] = {
            "filepath": file_path,
            "package_name": package_name,
        }

        if source.endswith("toml"):
            if source.startswith("pyproject"):
                sources.append(PyprojectSource(**source_kwargs))
            else:
                sources.append(TomlSource(**source_kwargs))

        if source.endswith("ini"):
            sources.append(IniSource(**source_kwargs))

    return sources


def deep_merge(destination: Dict[Any, Any], source: Dict[Any, Any]) -> Dict[Any, Any]:
    """Recursively updates the destination dictionary.

    Usage example:
    >>> a = { 'first' : { 'all_rows' : { 'pass' : 'dog', 'number' : '1' } } }
    >>> b = { 'first' : { 'all_rows' : { 'fail' : 'cat', 'number' : '5' } } }
    >>> deep_merge(a, b) == {
    ...     "first": {"all_rows": {"pass": "dog", "fail": "cat", "number": "5"}}
    ... }

    Note that the arguments may be modified!

    Based on https://stackoverflow.com/a/20666342

    Args:
        destination: A dictionary to be merged into. This will be updated in place.
        source: The dictionary supplying data

    Returns:
        The updated destination dictionary.

    Raises:
        RuntimeError: A dict cannot be merged on top of a non-dict.
            For example, the following would fail:
            `deep_merge({"foo": "bar"}, {"foo": {"baz": "qux"}})`
    """
    for key, src_value in source.items():
        if isinstance(src_value, dict):
            # get node or create one
            dest_node = destination.setdefault(key, {})
            if not isinstance(dest_node, dict):
                raise RuntimeError(
                    f"Cannot merge dict '{src_value}' into type '{type(dest_node)}'"
                )
            deep_merge(dest_node, src_value)
        else:
            destination[key] = src_value

    return destination



"""Tests for the `PyprojectSource` class."""

from pathlib import Path
from typing import Callable

from maison.config_sources.pyproject_source import PyprojectSource


class TestToDict:
    """Tests for the `to_dict` method."""

    def test_success(self, create_pyproject_toml: Callable[..., Path]) -> None:
        pyproject_path = create_pyproject_toml()

        pyproject_source = PyprojectSource(filepath=pyproject_path, package_name="foo")

        assert pyproject_source.to_dict() == {"bar": "baz"}

    def test_unrecognised_section_name(
        self, create_pyproject_toml: Callable[..., Path]
    ) -> None:
        """An empty dict is returned if the package name is not found"""
        pyproject_path = create_pyproject_toml(section_name="foo")

        pyproject_source = PyprojectSource(filepath=pyproject_path, package_name="bar")

        assert pyproject_source.to_dict() == {}

    def test_unrecognised_format(self, create_toml: Callable[..., Path]) -> None:
        """An unrecognised format or pyproject.toml returns an empty dict"""
        pyproject_path = create_toml(filename="pyproject.toml", content={"foo": "bar"})

        pyproject_source = PyprojectSource(filepath=pyproject_path, package_name="baz")

        assert pyproject_source.to_dict() == {}



"""Module to hold the `UserConfig` class definition."""

from functools import reduce
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Protocol
from typing import Type
from typing import Union

from maison.errors import NoSchemaError
from maison.utils import _collect_configs
from maison.utils import deep_merge


class _IsSchema(Protocol):
    """Protocol for config schemas."""

    def model_dump(self) -> Dict[Any, Any]:
        """Convert the validated config to a dict."""


class UserConfig:
    """Model the user configuration."""

    def __init__(
        self,
        package_name: str,
        starting_path: Optional[Path] = None,
        source_files: Optional[List[str]] = None,
        schema: Optional[Type[_IsSchema]] = None,
        merge_configs: bool = False,
    ) -> None:
        """Initialize the config.

        Args:
            package_name: the name of the package, to be used to find the right section
                in the config file
            starting_path: an optional starting path to start the search for config
                file
            source_files: an optional list of source config filenames or absolute paths
                to search for. If none is provided then `pyproject.toml` will be used.
            schema: an optional `pydantic` model to define the config schema
            merge_configs: an optional boolean to determine whether configs should be
                merged if multiple are found
        """
        self.source_files = source_files or ["pyproject.toml"]
        self.merge_configs = merge_configs
        self._sources = _collect_configs(
            package_name=package_name,
            source_files=self.source_files,
            starting_path=starting_path,
        )
        self._schema = schema
        self._values = self._generate_config_dict()

    def __str__(self) -> str:
        """Return the __str__.

        Returns:
            the string representation
        """
        return f"<class '{self.__class__.__name__}'>"

    @property
    def values(self) -> Dict[str, Any]:
        """Return the user's configuration values.

        Returns:
            the user's configuration values
        """
        return self._values

    @values.setter
    def values(self, values: Dict[str, Any]) -> None:
        """Set the user's configuration values."""
        self._values = values

    @property
    def discovered_paths(self) -> List[Path]:
        """Return a list of the paths to the config sources found on the filesystem.

        Returns:
            a list of the paths to the config sources
        """
        return [source.filepath for source in self._sources]

    @property
    def path(self) -> Optional[Union[Path, List[Path]]]:
        """Return the path to the selected config source.

        Returns:
            `None` is no config sources have been found, a list of the found config
            sources if `merge_configs` is `True`, or the path to the active config
            source if `False`
        """
        if len(self._sources) == 0:
            return None

        if self.merge_configs:
            return self.discovered_paths

        return self.discovered_paths[0]

    @property
    def schema(self) -> Optional[Type[_IsSchema]]:
        """Return the schema.

        Returns:
            the schema
        """
        return self._schema

    @schema.setter
    def schema(self, schema: Type[_IsSchema]) -> None:
        """Set the schema."""
        self._schema = schema

    def validate(
        self,
        schema: Optional[Type[_IsSchema]] = None,
        use_schema_values: bool = True,
    ) -> Dict[str, Any]:
        """Validate the configuration.

        Warning:
            Using this method with `use_schema_values` set to `True` will cast values to
            whatever is defined in the schema. For example, for the following schema:

                class Schema(ConfigSchema):
                    foo: str

            Validating a config with:

                {"foo": 1}

            Will result in:

                {"foo": "1"}

        Args:
            schema: an optional class that follows the `IsSchema` protocol that
                defines the schema. This takes precedence over a schema provided at
                object instantiation.
            use_schema_values: an optional boolean to indicate whether the result
                of passing the config through the schema should overwrite the existing
                config values, meaning values are cast to types defined in the schema as
                described above, and default values defined in the schema are used.

        Returns:
            the config values

        Raises:
            NoSchemaError: when validation is attempted but no schema has been provided
        """
        selected_schema: Union[Type[_IsSchema], None] = schema or self.schema

        if not selected_schema:
            raise NoSchemaError

        validated_schema = selected_schema(**self.values)

        if use_schema_values:
            self.values = validated_schema.model_dump()

        return self.values

    def _generate_config_dict(self) -> Dict[str, Any]:
        """Generate the config dict.

        If `merge_configs` is set to `False` then we use the first config. If `True`
        then the dicts of the sources are merged from right to left.

        Returns:
            the config dict
        """
        if len(self._sources) == 0:
            return {}

        if not self.merge_configs:
            return self._sources[0].to_dict()

        source_dicts = (source.to_dict() for source in self._sources)
        return reduce(lambda a, b: deep_merge(a, b), source_dicts)



"""Module to hold the `TomlSource` class definition."""

from functools import lru_cache
from typing import Any
from typing import Dict

import toml

from ..errors import BadTomlError
from .base_source import BaseSource


class TomlSource(BaseSource):
    """Class to represent a `.toml` config source."""

    def to_dict(self) -> Dict[Any, Any]:
        """Convert the source config file to a dict.

        Returns:
            a dict of the config options and values
        """
        return self._load_file()

    @lru_cache
    def _load_file(self) -> Dict[Any, Any]:
        """Load the `.toml` file.

        Returns:
            the `.toml` source converted to a `dict`

        Raises:
            BadTomlError: If toml cannot be parsed
        """
        try:
            return dict(toml.load(self.filepath))
        except toml.decoder.TomlDecodeError as exc:
            raise BadTomlError(
                f"Error trying to load toml file '{self.filepath}'"
            ) from exc



"""Tests for the `utils` module."""

from pathlib import Path
from typing import Callable
from typing import Dict
from unittest.mock import MagicMock
from unittest.mock import patch

import pytest

from maison.utils import deep_merge
from maison.utils import get_file_path
from maison.utils import path_contains_file


class TestContainsFile:
    """Tests for the `contains_file` function"""

    def test_found(self, create_tmp_file: Callable[..., Path]) -> None:
        """Return `True` if the path contains the file"""
        path = create_tmp_file(filename="file.txt")

        result = path_contains_file(path=path.parent, filename="file.txt")

        assert result is True

    def test_not_found(self, create_tmp_file: Callable[..., Path]) -> None:
        """Return `False` if the path does not contain the file"""
        path = create_tmp_file(filename="file.txt")

        result = path_contains_file(path=path.parent, filename="other.txt")

        assert result is False


class TestGetFilePath:
    """Tests for the `get_file_path`"""

    @patch("maison.utils.Path", autospec=True)
    def test_in_current_directory(
        self, mock_path: MagicMock, create_tmp_file: Callable[..., Path]
    ) -> None:
        """The path to a file is returned."""
        mock_path.return_value.expanduser.return_value.is_absolute.return_value = False

        path_to_file = create_tmp_file(filename="file.txt")
        mock_path.cwd.return_value = path_to_file.parent

        result = get_file_path(filename="file.txt")

        assert result == path_to_file

    def test_in_parent_directory(self, create_tmp_file: Callable[..., Path]) -> None:
        """The path to a file in a parent directory is returned."""
        path_to_file = create_tmp_file(filename="file.txt")
        sub_dir = path_to_file / "sub"

        result = get_file_path(filename="file.txt", starting_path=sub_dir)

        assert result == path_to_file

    def test_not_found(self) -> None:
        """If the file isn't found in the tree then return a `None`"""
        result = get_file_path(filename="file.txt", starting_path=Path("/nowhere"))

        assert result is None

    def test_with_given_path(self, create_tmp_file: Callable[..., Path]) -> None:
        """A `starting_path` can be used to initiate the starting search directory"""
        path_to_file = create_tmp_file(filename="file.txt")

        result = get_file_path(filename="file.txt", starting_path=path_to_file)

        assert result == path_to_file

    def test_absolute_path(self, create_tmp_file: Callable[..., Path]) -> None:
        """An absolute path to an existing file is returned"""
        path_to_file = create_tmp_file(filename="file.txt")

        result = get_file_path(filename=str(path_to_file))

        assert result == path_to_file

    def test_absolute_path_not_exist(self) -> None:
        """If the absolute path doesn't exist return a `None`"""
        result = get_file_path(filename="~/xxxx/yyyy/doesnotexist.xyz")

        assert result is None


class TestDeepMerge:
    """Tests for the `deep_merge` function."""

    @pytest.mark.parametrize(
        ("a", "b", "expected"),
        [
            pytest.param(
                {1: 2, 3: 4},
                {3: 5, 6: 7},
                {1: 2, 3: 5, 6: 7},
                id="simple",
            ),
            pytest.param(
                {1: 2, 3: {4: 5, 6: 7}},
                {3: {6: 8, 9: 10}, 11: 12},
                {1: 2, 3: {4: 5, 6: 8, 9: 10}, 11: 12},
                id="nested",
            ),
        ],
    )
    def test_success(
        self,
        a: Dict[int, int],
        b: Dict[int, int],
        expected: Dict[int, int],
    ) -> None:
        assert deep_merge(a, b) == expected
        assert a == expected

    def test_incompatible_dicts(self) -> None:
        """Trying to merge incompatible dicts returns an error"""
        dict_a = {1: 2, 2: 5}
        dict_b = {1: {3: 4}}

        with pytest.raises(RuntimeError):
            deep_merge(dict_a, dict_b)



"""Module to define custom errors."""


class NoSchemaError(Exception):
    """Raised when validation is attempted but no schema has been provided."""


class BadTomlError(Exception):
    """Raised when loading from an invalid toml source is attempted."""



"""Module to hold the `PyprojectSource` class definition."""

from typing import Any
from typing import Dict

from .toml_source import TomlSource


class PyprojectSource(TomlSource):
    """Class to represent a `pyproject.toml` config source."""

    def to_dict(self) -> Dict[Any, Any]:
        """Convert the package `pyproject.toml` section to a dict.

        Relies on the convention that config related to package `acme` will be
        located under a `[tool.acme]` section in `pyproject.toml`

        Returns:
            a dict of the config options and values
        """
        return dict(self._load_file().get("tool", {}).get(self.package_name, {}))



"""Tests for the `BaseSource` class."""

from pathlib import Path
from typing import Any
from typing import Callable
from typing import Dict

from maison.config_sources.base_source import BaseSource


class ConcreteSource(BaseSource):
    """Concretion of `BaseSource` for testing purposes"""

    def to_dict(self) -> Dict[Any, Any]:
        """Return a dict."""
        return {}


class TestRepr:
    """Tests for the `__repr__` method."""

    def test_contains_path(self) -> None:
        source = ConcreteSource(filepath=Path("~/file.txt"), package_name="acme")

        assert "file.txt" in repr(source)
        assert str(source) == repr(source)


class TestFilename:
    """Tests for the `filename` property."""

    def test_success(self, create_tmp_file: Callable[..., Path]) -> None:
        path_to_file = create_tmp_file(filename="file.txt")

        source = ConcreteSource(filepath=path_to_file, package_name="acme")

        assert source.filename == "file.txt"



"""Tests for the `IniSource` class."""

from pathlib import Path
from typing import Callable

from maison.config_sources.ini_source import IniSource


class TestToDict:
    """Tests for the `to_dict` method."""

    def test_success(self, create_tmp_file: Callable[..., Path]) -> None:
        """A `.ini` file is converted to a `dict`"""
        ini_file = """
[section 1]
option_1 = value_1

[section 2]
option_2 = value_2
        """
        ini_path = create_tmp_file(content=ini_file, filename="foo.ini")

        toml_source = IniSource(filepath=ini_path, package_name="acme")

        assert toml_source.to_dict() == {
            "section 1": {"option_1": "value_1"},
            "section 2": {"option_2": "value_2"},
        }

    def test_empty_file(self, create_tmp_file: Callable[..., Path]) -> None:
        """Empty `.ini` returns an empty dict"""
        ini_path = create_tmp_file(filename="foo.ini")

        toml_source = IniSource(filepath=ini_path, package_name="acme")

        assert toml_source.to_dict() == {}



"""Tests for the `Config` classes."""

from pathlib import Path
from typing import Callable

import pytest
from pydantic import BaseModel
from pydantic import ValidationError

from maison.config import UserConfig
from maison.errors import NoSchemaError


class TestUserConfig:
    """Tests for the `UserConfig` class."""

    def test_str(self, create_tmp_file: Callable[..., Path]) -> None:
        pyproject_path = create_tmp_file(filename="pyproject.toml")

        config = UserConfig(package_name="foo", starting_path=pyproject_path)

        assert str(config) == "<class 'UserConfig'>"


class TestDictObject:
    """Tests to ensure that the config is accessible as a dict."""

    def test_valid_pyproject(self, create_pyproject_toml: Callable[..., Path]) -> None:
        """A valid pyproject is parsed to a dict object."""
        pyproject_path = create_pyproject_toml()

        config = UserConfig(package_name="foo", starting_path=pyproject_path)

        assert config.values == {"bar": "baz"}


class TestSourceFiles:
    """Tests for the `source_files` init argument."""

    def test_not_found(self) -> None:
        """Non existent source files are handled."""
        config = UserConfig(package_name="foo", source_files=["foo"])

        assert config.path is None
        assert config.values == {}

    def test_unrecognised_file_extension(
        self,
        create_tmp_file: Callable[..., Path],
    ) -> None:
        """Unrecognised source file extensions are handled."""
        source_path = create_tmp_file(filename="foo.txt")
        config = UserConfig(
            package_name="foo",
            source_files=["foo.txt"],
            starting_path=source_path,
        )

        assert config.path is None
        assert config.values == {}

    def test_single_valid_toml_source(self, create_toml: Callable[..., Path]) -> None:
        """Toml files other than pyproject.toml files are handled."""
        source_path = create_toml(filename="another.toml", content={"bar": "baz"})

        config = UserConfig(
            package_name="foo",
            starting_path=source_path,
            source_files=["another.toml"],
        )

        assert config.path == source_path
        assert config.values["bar"] == "baz"

    def test_multiple_valid_toml_sources(
        self,
        create_pyproject_toml: Callable[..., Path],
        create_toml: Callable[..., Path],
    ) -> None:
        """When there are multiple sources, the first one is used"""
        source_path_1 = create_toml(filename="another.toml", content={"bar": "baz"})

        source_path_2 = create_pyproject_toml(
            section_name="oof", content={"rab": "zab"}
        )

        config = UserConfig(
            package_name="foo",
            starting_path=source_path_2,
            source_files=["another.toml", "pyproject.toml"],
        )

        assert config.discovered_paths == [source_path_1, source_path_2]
        assert config.values["bar"] == "baz"

    def test_absolute_path(self, create_tmp_file: Callable[..., Path]) -> None:
        """Source files can be found using absolute paths"""
        path = create_tmp_file(filename="acme.ini")

        config = UserConfig(
            package_name="foo",
            source_files=[str(path)],
        )

        assert config.discovered_paths == [path]

    def test_absolute_path_not_exist(
        self,
        create_pyproject_toml: Callable[..., Path],
    ) -> None:
        """Non existent absolute paths are handled."""
        pyproject_path = create_pyproject_toml()

        config = UserConfig(
            package_name="foo",
            source_files=["~/.config/acme.ini", "pyproject.toml"],
            starting_path=pyproject_path,
        )

        assert config.discovered_paths == [pyproject_path]


class TestIniFiles:
    """Tests for handling x.ini config files."""

    def test_valid_ini_file(self, create_tmp_file: Callable[..., Path]) -> None:
        ini_file = """
[section 1]
option_1 = value_1

[section 2]
option_2 = value_2
        """
        source_path = create_tmp_file(content=ini_file, filename="foo.ini")
        config = UserConfig(
            package_name="foo",
            starting_path=source_path,
            source_files=["foo.ini"],
        )

        assert config.discovered_paths == [source_path]
        assert config.values == {
            "section 1": {"option_1": "value_1"},
            "section 2": {"option_2": "value_2"},
        }


class TestValidation:
    """Tests for schema validation."""

    def test_no_schema(self) -> None:
        config = UserConfig(package_name="acme", starting_path=Path("/"))

        assert config.values == {}

        with pytest.raises(NoSchemaError):
            config.validate()

    def test_one_schema_with_valid_config(
        self,
        create_pyproject_toml: Callable[..., Path],
    ) -> None:
        """The config is validated with a given schema."""

        class Schema(BaseModel):
            """Defines schema."""

            bar: str

        pyproject_path = create_pyproject_toml()
        config = UserConfig(
            package_name="foo",
            starting_path=pyproject_path,
            schema=Schema,
        )

        config.validate()

        assert config.values["bar"] == "baz"

    def test_one_schema_injected_at_validation(
        self,
        create_pyproject_toml: Callable[..., Path],
    ) -> None:
        """Schemas supplied as an argument are used"""

        class Schema(BaseModel):
            """Defines schema."""

            bar: str

        pyproject_path = create_pyproject_toml()
        config = UserConfig(
            package_name="foo",
            starting_path=pyproject_path,
        )

        config.validate(schema=Schema)

        assert config.values["bar"] == "baz"

    def test_use_schema_values(
        self,
        create_pyproject_toml: Callable[..., Path],
    ) -> None:
        """Config values can be cast to the validated values."""

        class Schema(BaseModel, coerce_numbers_to_str=True):
            """Defines schema."""

            bar: str
            other: str = "hello"

        pyproject_path = create_pyproject_toml(content={"bar": 1})
        config = UserConfig(
            package_name="foo",
            starting_path=pyproject_path,
            schema=Schema,
        )

        config.validate()

        assert config.values["bar"] == "1"
        assert config.values["other"] == "hello"

    def test_not_use_schema_values(
        self,
        create_pyproject_toml: Callable[..., Path],
    ) -> None:
        """If `use_schema_values` is set to False then don't use validated values."""

        class Schema(BaseModel, coerce_numbers_to_str=True):
            """Defines schema."""

            bar: str
            other: str = "hello"

        pyproject_path = create_pyproject_toml(content={"bar": 1})
        config = UserConfig(
            package_name="foo",
            starting_path=pyproject_path,
            schema=Schema,
        )

        config.validate(use_schema_values=False)

        assert config.values["bar"] == 1
        assert "other" not in config.values

    def test_schema_override(
        self,
        create_pyproject_toml: Callable[..., Path],
    ) -> None:
        """Schemas given as an argument are preferred"""

        class InitSchema(BaseModel):
            """Defines schema for 1."""

            bar: str = "schema_1"

        class ArgumentSchema(BaseModel):
            """Defines schema for 2."""

            bar: str = "schema_2"

        pyproject_path = create_pyproject_toml(content={"baz": "baz"})
        config = UserConfig(
            package_name="foo",
            starting_path=pyproject_path,
            schema=InitSchema,
        )

        config.validate(schema=ArgumentSchema)

        assert config.values["bar"] == "schema_2"

    def test_invalid_configuration(
        self,
        create_pyproject_toml: Callable[..., Path],
    ) -> None:
        """Validation errors are raised when config fails validation."""

        class Schema(BaseModel):
            """Defines schema."""

            bar: str

        pyproject_path = create_pyproject_toml(content={"baz": "baz"})
        config = UserConfig(
            package_name="foo",
            starting_path=pyproject_path,
            schema=Schema,
        )

        with pytest.raises(ValidationError):
            config.validate()

    def test_setter(self) -> None:
        """Schemas can be set using the setter."""

        class Schema(BaseModel):
            """Defines schema."""

        config = UserConfig(package_name="foo")

        assert config.schema is None

        config.schema = Schema

        assert config.schema is Schema


class TestMergeConfig:
    """Tests for the merging of multiple config sources."""

    def test_no_overwrites(
        self,
        create_toml: Callable[..., Path],
        create_tmp_file: Callable[..., Path],
        create_pyproject_toml: Callable[..., Path],
    ) -> None:
        """Configs without overlapping values are merged."""
        config_1_path = create_toml(filename="config.toml", content={"option_1": True})
        ini_file = """
[foo]
option_2 = true
        """
        config_2_path = create_tmp_file(filename="config.ini", content=ini_file)
        pyproject_path = create_pyproject_toml(content={"option_3": True})

        config = UserConfig(
            package_name="foo",
            source_files=[str(config_1_path), str(config_2_path), "pyproject.toml"],
            starting_path=pyproject_path,
            merge_configs=True,
        )

        assert config.path == [config_1_path, config_2_path, pyproject_path]
        assert config.values == {
            "option_1": True,
            "foo": {
                "option_2": "true",
            },
            "option_3": True,
        }

    def test_overwrites(
        self,
        create_toml: Callable[..., Path],
        create_pyproject_toml: Callable[..., Path],
    ) -> None:
        """Configs with overlapping values are merged."""
        config_1_path = create_toml(
            filename="config_1.toml", content={"option": "config_1"}
        )
        config_2_path = create_toml(
            filename="config_2.toml", content={"option": "config_2"}
        )
        pyproject_path = create_pyproject_toml(content={"option": "config_3"})

        config = UserConfig(
            package_name="foo",
            source_files=[str(config_1_path), str(config_2_path), "pyproject.toml"],
            starting_path=pyproject_path,
            merge_configs=True,
        )

        assert config.values == {
            "option": "config_3",
        }

    def test_nested(
        self,
        create_toml: Callable[..., Path],
        create_pyproject_toml: Callable[..., Path],
    ) -> None:
        """Configs with nested overlapping values are deep merged."""
        config_1_path = create_toml(
            filename="config_1.toml", content={"option": {"nested_1": "config_1"}}
        )
        config_2_path = create_toml(
            filename="config_2.toml", content={"option": {"nested_2": "config_2"}}
        )
        pyproject_path = create_pyproject_toml(
            content={"option": {"nested_2": "config_3"}}
        )

        config = UserConfig(
            package_name="foo",
            source_files=[str(config_1_path), str(config_2_path), "pyproject.toml"],
            starting_path=pyproject_path,
            merge_configs=True,
        )

        assert config.values == {
            "option": {"nested_1": "config_1", "nested_2": "config_3"},
        }
